/*
 * Copyright (C) 2012   Pierre Habouzit <pierre.habouzit@intersec.com>
 * Copyright (C) 2012   Intersec SAS
 *
 * This file implements the Linux Pthread Workqueue Regulator usperspace.
 *
 * The Linux Pthread Workqueue Regulator is free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser
 * General Public License as published by the Free Software Foundation,
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * The Linux Pthread Workqueue Regulator is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with The Linux Pthread Workqueue Regultaor.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <stddef.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <unistd.h>

#include "../kernel/pwqr.h"

#define INPOOL          0
#define PARKED          1
#define PARKED_MAX     16
#define THREADS_MAX   128 /* something dynamic is probably better */

/* Compiler stuff {{{ */

#if defined(__x86_64__)
#  define mb()          asm volatile("mfence":::"memory")
#elif defined(__i386__)
#  define mb()          asm volatile("lock; addl $0,0(%%esp)":::"memory")
#else
#  error unsupported architecture
#endif

#define container_of(obj, type_t, member) \
      ({ const typeof(((type_t *)0)->member) *__mptr = (obj); \
         ((type_t *)((char *)__mptr) - offsetof(type_t, member)); })
#define likely(expr)     __builtin_expect(!!(expr), 1)
#define unlikely(expr)   __builtin_expect((expr), 0)

#define barrier()          asm volatile("":::"memory")
#define access_once(x)     (*(volatile typeof(x) *)&(x))

#define atomic_fetch_and_add(ptr, n) __sync_fetch_and_add(ptr, n)
#define atomic_add(ptr, n)           (void)atomic_fetch_and_add(ptr, n)
#define atomic_fetch_and_sub(ptr, n) __sync_fetch_and_sub(ptr, n)
#define atomic_sub(ptr, n)           (void)atomic_fetch_and_sub(ptr, n)
#define atomic_xchg(ptr, v)          ((typeof(*(ptr)))__sync_lock_test_and_set(ptr, v))

#define ALWAYS_INLINE     __attribute__((always_inline)) inline

/* }}} */
/* pwqr wrapping {{{ */

static int pwqr_create(int flags)
{
    if (flags & ~PWQR_FL__SET) {
        errno = -EINVAL;
        return -1;
    }
    return open("/dev/"PWQR_DEVICE_NAME, O_RDWR | flags);
}

static int pwqr_ctl(int fd, int op, int val, void *uaddr)
{
    struct pwqr_ioc_wait wait;

    switch (op) {
      case PWQR_CTL_GET_CONC:
      case PWQR_CTL_REGISTER:
      case PWQR_CTL_UNREGISTER:
      case PWQR_CTL_PARK:
        return ioctl(fd, op);
      case PWQR_CTL_SET_CONC:
      case PWQR_CTL_WAKE:
      case PWQR_CTL_WAKE_OC:
        return ioctl(fd, op, val);
      case PWQR_CTL_WAIT:
        wait.pwqr_ticket = val;
        wait.pwqr_uaddr  = uaddr;
        return ioctl(fd, op, &wait);
      default:
        errno = EINVAL;
        return -1;
    }
}

/* }}} */

static struct {
    int fd;
    /* atomic */ int spinlock;
    /* atomic */ int nthreads;
    /* atomic */ int waiters;
    /* atomic */ int parked;
    /* atomic */ int overcommit_count;

    union {
        struct {
#if __BYTE_ORDER == __LITTLE_ENDIAN
            /* atomic */ uint32_t lo;
            /* atomic */ uint32_t hi;
#elif __BYTE_ORDER == __BIG_ENDIAN
            /* atomic */ uint32_t hi;
            /* atomic */ uint32_t lo;
#else
#error  Unsupported endianness
#endif
        };
        uint64_t ticket;
    };
} pwqr_g = {
    .fd = -1,
};

static ALWAYS_INLINE void pwqr_lock(void)
{
    while (unlikely(__sync_lock_test_and_set(&pwqr_g.spinlock, 1)))
        sched_yield();
}

static ALWAYS_INLINE void pwqr_unlock(void)
{
    __sync_lock_release(&pwqr_g.spinlock);
}

static ALWAYS_INLINE uint64_t pwqr_get_ticket(void)
{
    mb();
    return access_once(pwqr_g.ticket);
}

static ALWAYS_INLINE void pwqr_signal_n(int n)
{
#if ULONG_MAX == UINT32_MAX
    /*
     * We don't know how slow things are between a pwqr_get() and a
     * pwqr_wait. Hence it's not safe to assume the "count" will never
     * rotate fully between the two.
     *
     * In 64bits mode, well, we have 64-bits atomic increments and all is
     * fine. In 32bits mode, we increment the high word manually every 2^24
     * low-word increment.
     *
     * This assumes that at least one of the 256 threads that will try to
     * perform the "hi" increment won't be stalled between deciding the modulo
     * is zero and the increment itself for an almost full cycle (2^32 - 2^24)
     * of the low word counter.
     */
    if (unlikely(atomic_fetch_and_add(&pwqr_g.lo, 1) % (1U << 24) == 0))
        atomic_add(&pwqr_g.hi, 1);
#else
    atomic_add(&pwqr_g.ticket, 1);
#endif
    if (atomic_fetch_and_add(&pwqr_g.waiters, 0))
        pwqr_ctl(pwqr_g.fd, PWQR_CTL_WAKE, n, NULL);
}
static ALWAYS_INLINE void pwqr_signal(void)
{
    pwqr_signal_n(1);
}

static ALWAYS_INLINE void pwqr_signal_relaxed_n(int n)
{
    if (access_once(pwqr_g.waiters))
        pwqr_signal_n(n);
}
static ALWAYS_INLINE void pwqr_signal_relaxed(void)
{
    pwqr_signal_relaxed_n(1);
}

static int pwqr_fd_read_overcommit(int fd)
{
    int buf;

    if (read(fd, &buf, sizeof(buf)) == sizeof(buf))
        return buf;
    return 0;
}

__attribute__((cold))
static int pwqr_fd_overcommit_check(void)
{
    if (atomic_xchg(&pwqr_g.overcommit_count, 0)) {
        int oc = pwqr_fd_read_overcommit(pwqr_g.fd);

        if (oc) {
            access_once(pwqr_g.overcommit_count) = (oc - 1);
            return 1;
        }
    }
    return 0;
}

__attribute__((unused))
static void pwqr_overcommit_poll_loop(void)
{
    struct pollfd pfd = {
        .fd     = pwqr_g.fd,
        .events = POLLIN,
    };

    for (;;) {
        pfd.revents = 0;
        if (poll(&pfd, 1, -1) >= 0) {
            if (pfd.revents & POLLIN) {
                access_once(pwqr_g.overcommit_count) =
                    pwqr_fd_read_overcommit(pwqr_g.fd);
            }
            if (pfd.revents & (POLLHUP | POLLERR))
                return;
        } else if (errno != EINTR && errno != EAGAIN) {
            return;
        }
    }
}


/* returns:
   - INPOOL if we were woken to run jobs.
   - PARKED if we were woken to be parked (possibly consume overcommit stuff)
 */
static void pwqr_do_wait_cleanup(void *arg)
{
    atomic_sub(&pwqr_g.waiters, 1);
}
static int pwqr_do_wait(uint64_t ticket)
{
    int canceltype, rc;

    mb();
#if ULONG_MAX == UINT32_MAX
    if ((unsigned)(ticket >> 32) != access_once(pwqr_g.hi))
        return INPOOL;
#else
    if (ticket != access_once(pwqr_g.ticket))
        return INPOOL;
#endif

    atomic_add(&pwqr_g.waiters, 1);
    pthread_cleanup_push(&pwqr_do_wait_cleanup, NULL);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &canceltype);

    rc = pwqr_ctl(pwqr_g.fd, PWQR_CTL_WAIT, ticket, &pwqr_g.lo);
    if (rc < 0) {
        if (errno == EINTR) {
            rc = INPOOL;
        } else if (errno == EDQUOT) {
            rc = PARKED;
        } else {
            assert (errno == EBADFD);
        }
    }

    pthread_setcanceltype(canceltype, NULL);
    pthread_cleanup_pop(1);
}

/* returns INPOOL if the thread is to run some jobs or go to WAIT
 * returns -1 if the pwqr_g.fd is broken
 */
static void pwqr_do_park_cleanup(void *arg)
{
    atomic_sub(&pwqr_g.parked, 1);
}
static void pwqr_spawn_thread(int initial_state, int count);
static int pwqr_do_park(void)
{
    int canceltype, rc = -1;

    if (atomic_fetch_and_add(&pwqr_g.parked, 1) > PARKED_MAX) {
        atomic_sub(&pwqr_g.parked, 1);
        return -1;
    }

    pthread_cleanup_push(&pwqr_do_park_cleanup, NULL);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &canceltype);

    for (;;) {
#ifdef PSEUDO_CODE
        while ((job = find_some_job(OC_JOB)))
            run_job(job);
#endif

        if ((rc = pwqr_ctl(pwqr_g.fd, PWQR_CTL_PARK, 0, NULL)) == 0) {
            access_once(pwqr_g.overcommit_count) = 0;
            if (atomic_fetch_and_sub(&pwqr_g.parked, 1) == 0) {
                pwqr_spawn_thread(PARKED, 1);
            }
            break;
        }

        if (rc < 0) {
            if (errno == EBADFD) {
                atomic_sub(&pwqr_g.parked, 1);
                return -1;
            }
            assert (errno == EINTR || errno == EDQUOT);
            continue;
        }
    }

    pthread_setcanceltype(canceltype, NULL);
    pthread_cleanup_pop(0);
    return rc;
}

static void pwqr_main_cleanup(void *arg)
{
    atomic_sub(&pwqr_g.nthreads, 1);
}
static void *pwqr_main(void *arg)
{
    int state = (long)arg;

    if (atomic_fetch_and_add(&pwqr_g.nthreads, 1) > THREADS_MAX)
        goto out;

    pthread_cleanup_push(&pwqr_main_cleanup, NULL);
    if (pwqr_ctl(pwqr_g.fd, PWQR_CTL_REGISTER, 0, NULL) < 0)
        goto out;

    do {
        uint64_t ticket;

      park:
        if (state == PARKED && (state = pwqr_do_park()) < 0)
            break;
        for (;;) {
            if (unlikely(pwqr_g.overcommit_count)) {
                if (pwqr_fd_overcommit_check()) {
                    state = PARKED;
                    goto park;
                }
            }
#ifdef PSEUDO_CODE
            job = find_some_job(ANY_JOB);
            if (!job)
                break;
            run_job(job);
#endif
        }
        ticket = pwqr_get_ticket();
#ifdef PSEUDO_CODE
        if ((job = find_some_job(ANY_JOB))) {
            run_job(job);
            continue;
        }
#endif
        state = pwqr_do_wait(ticket);
    } while (state >= 0);

  out:
    pthread_cleanup_pop(1);
    return NULL;
}

static void pwqr_spawn_thread(int initial_state, int count)
{
    pthread_t thr;
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    while (count-- > 0)
        pthread_create(&thr, &attr, pwqr_main, (void *)(long)initial_state);
    pthread_attr_destroy(&attr);
}

__attribute__((cold, noinline, unused))
static int _pthread_workqueue_init_np(void)
{
    int rc = 0, fd, n;

    pwqr_lock();
    if (pwqr_g.fd >= 0)
        goto out;

    fd = pwqr_create(0);
    if (fd < 0) {
        rc = -1;
        goto out;
    }

    pwqr_g.fd = fd;
    n = pwqr_ctl(pwqr_g.fd, PWQR_CTL_GET_CONC, 0, NULL) + 4;
    pwqr_spawn_thread(INPOOL, n);
    pwqr_spawn_thread(PARKED, 4);

  out:
    pwqr_unlock();
    return rc;
}
